/*

------------------------------------------------------------------------------

A license is hereby granted to reproduce this software source code and
to create executable versions from this source code for personal,
non-commercial use.  The copyright notice included with the software
must be maintained in all copies produced.

THIS PROGRAM IS PROVIDED "AS IS". THE AUTHOR PROVIDES NO WARRANTIES
WHATSOEVER, EXPRESSED OR IMPLIED, INCLUDING WARRANTIES OF
MERCHANTABILITY, TITLE, OR FITNESS FOR ANY PARTICULAR PURPOSE.  THE
AUTHOR DOES NOT WARRANT THAT USE OF THIS PROGRAM DOES NOT INFRINGE THE
INTELLECTUAL PROPERTY RIGHTS OF ANY THIRD PARTY IN ANY COUNTRY.

Copyright (c) 1994-2007, John Conover, All Rights Reserved.

Comments and/or bug reports should be addressed to:

    john@email.johncon.com (John Conover)

------------------------------------------------------------------------------

ucspi-daemon.c is for providing interoperability and transaction
processing for the tcpserver and tcpclient programs at
http://cr.yp.to/ucspi-tcp.html. Only text transactions are permitted,
(the program uses the EOT = ^D = end of transmission/transaction
character for EOF and transaction concurrency control across the
link, which is not compatible with binary data.)

For example:

    tcpserver -v 0 23456 ucspi-daemon -s rw -i file1 -o file2
    tcpclient -v 0 23456 ucspi-daemon -c wr -i file3 -o file4

would establish a connection on localhost, port 23456, and transfer
file3 to file2, then transfer file1 to file4, as directed by the -c
wr, (write to the network, then read,) and -s rw, (read from the
network, then write.) The files can be named pipes for accessing
program i/o across a network.

The sequence of r's and w's can be any length, but the server -s
argument must be a "mirror" of the client, i.e., r->w, and vice verse;
the r's and w's should alternate for flow/transaction control.

The output file is advisory write locked for the duration of the
transaction for concurrency control.

If arguments are supplied on the command line, the arguments will be
EOS = '\n' terminated, and sent to the link, instead of the input
file.

The -r option does not strip the EOT = ^D = end of
transmission/transaction character from the data sent to the output
file, (the default is to strip it,) for use by programs interoperating
over the link.

$Revision: 0.0 $
$Date: 2007/01/18 19:36:00 $
$Id: ucspi-daemon.c,v 0.0 2007/01/18 19:36:00 john Exp $
$Log: ucspi-daemon.c,v $
Revision 0.0  2007/01/18 19:36:00  john
Initial version


*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>

#define _POSIX_SOURCE 1 /* POSIX compliant source */

static char rcsid[] = "$Id: ucspi-daemon.c,v 0.0 2007/01/18 19:36:00 john Exp $"; /* program version */
static char copyright[] = "Copyright (c) 1994-2007, John Conover, All Rights Reserved"; /* the copyright banner */

#define BUFLEN BUFSIZ /* i/o buffer size */

#ifdef __STDC__

static const char *help_message[] = /* help message index array */

#else

static char *help_message[] = /* help message index array */

#endif

{
    "\n",
    "Ucspi tcpserver/tcpclient daemon\n",
    "Usage: ucspi-daemon [-c w|r|wr|rw ...] [-i infile] [-o outfile] [-r]\n",
    "                    [-s r|w|rw|wr ...] [-v] [argument(s)]\n",
    "    -c w|r|wr|rw ..., ucspi/tcpclient daemon read/write\n",
    "                      transaction sequence, (w);\n",
    "                      mutually exclusive with -s\n",
    "    -i infile, input filename, (stdin)\n",
    "    -o outfile, output filename, (stdout)\n",
    "    -r, do not remove EOT = ^D = end of transmission/transaction from\n",
    "        outfile\n",
    "    -s r|w|rw|wr ..., ucspi/tcpserver daemon read/write\n",
    "                      transaction sequence, (r);\n",
    "                      mutually exclusive with -c, (default)\n",
    "    -v, print the program's version information\n",
    "    argument(s), concatenate arguments, instead of reading infile\n"
};

#ifdef __STDC__

static const char *error_message[] = /* error message index array */

#else

static char *error_message[] = /* error message index array */

#endif

{
    "No error\n",
    "Error in program argument(s)\n",
    "Error opening input file\n",
    "Error closing input file\n",
    "Error reading input file\n",
    "Error opening output file\n",
    "Error closing output file\n",
    "Error writing to output file\n",
    "Error locking output file\n",
    "Error unlocking output file\n"
};

#define NOERROR 0 /* error values, one for each index in the error message array */
#define EARGS 1
#define EOPENIN 2
#define ECLOSEIN 3
#define ERDIN 4
#define EOPENOUT 5
#define ECLOSEOUT 6
#define EWROUT 7
#define EOLCK 8
#define EOULCK 9

#ifdef __STDC__

static int file_operation (int read_in, int write_out, int argc, char *argv[], int arg_ctr, char eot_op); /* file i/o operation */
static void print_message (int retval); /* print any error messages */

#else

static int file_operation (); /* file i/o operation */
static void print_message (); /* print any error messages */

#endif

#ifdef __STDC__

static const char *transaction_sequence = {"r"}, /* transaction sequence */
                  *transaction_ref; /* reference to transaction type */

#else

static char *transaction_sequence = {"r"}, /* transaction sequence */
            *transaction_ref; /* reference to transaction type */

#endif

static char EOT = (char) 0x04; /* EOT = ^D, end of transmission/transaction marker character */

#ifdef __STDC__

int main (int argc, char *argv[])

#else

int main (argc, argv)
int argc;
char *argv[];

#endif

{
    char *infile_name = (char *) 0, /* reference to input file name */
         *outfile_name = (char *) 0, /* reference to output file name */
         transaction; /* the current transaction */

    int retval = NOERROR, /* return value, assume no error */
        c, /* command line switch */
        arg_ctr, /* command line argument counter */
        add_eot = 0, /* add EOT = ^D = end of transmission/transaction marker character to output file, 0 = no, 1 = yes */
        server = 1, /* server flag, 0 = operating as client, 1 = operating as server */
        infile = STDIN_FILENO, /* input file descriptor */
        outfile = STDOUT_FILENO, /* output file descriptor */
        interin = STDIN_FILENO, /* intermediate input device descriptor */
        interout = STDOUT_FILENO; /* intermediate output device descriptor */

    struct flock outlock; /* output file locking structure */

    outlock.l_type = F_WRLCK; /* lock/unlock EOF of file for write */
    outlock.l_whence = (short int) SEEK_SET;
    outlock.l_start = (off_t) 0;
    outlock.l_len = (off_t) 0;

    while ((c = getopt (argc, argv, "c:i:o:prs:v")) != EOF) /* for each command line switch */
    {

        switch (c) /* which switch? */
        {

            case 'c': /* request for client daemon under tcpclient? */

                transaction_sequence = optarg; /* yes, save the client read/write transaction sequence */
                server = 0; /* reset the server flag, 0 = operating as client, 1 = operating as server */
                interin = 6; /* intermediate input device descriptor for ucspi/tcpclient */
                interout = 7; /* intermediate output device descriptor for ucspi/tcpclient */
                break;

            case 'i': /* request for input file name? */

                infile_name = optarg; /* yes, save the input file name */
                break;

            case 'o': /* request for output file name? */

                outfile_name = optarg; /* yes, save the output file name */
                break;

            case 'r': /* request to add EOT = ^D = end of transmission/transaction marker character to output file, 0 = no, 1 = yes */

                add_eot = 1; /* set add EOT = ^D = end of transmission/transaction marker character to output file, 0 = no, 1 = yes */
                break;

            case 's': /* request for server daemon under tcpserver? */

                transaction_sequence = optarg; /* yes, save the server read/write transaction sequence */
                server = 1; /* set the server flag, 0 = operating as client, 1 = operating as server */
                interin = STDIN_FILENO; /* intermediate input device descriptor */
                interout = STDOUT_FILENO; /* intermediate output device descriptor */
                break;

            case 'v':

                (void) printf ("%s\n", rcsid); /* print the version */
                (void) printf ("%s\n", copyright); /* print the copyright */
                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */

            default: /* illegal switch? */

                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */
                break;

        }

    }

    transaction_ref = transaction_sequence; /* reference to transaction type */

    while ((transaction = *transaction_ref) != '\0') /* the transaction sequence that will be followed: check for valid characters */
    {

        switch (transaction) /* which transaction? */
        {

            case 'r': /* read? */
            case 'R':

                break;

            case 'w': /* write? */
            case 'W':

                break;

            default: /* non-recognized transaction? */

                retval = EARGS; /* assume not enough arguments */
                break;

        }

        transaction_ref ++; /* increment the transaction sequence that will be followed */
    }

    if (retval == NOERROR) /* any errors? */
    {
        arg_ctr = optind; /* save remaining command line arguments, if any */
        retval = EOPENIN; /* assume error opening input file */

        if ((infile = (infile_name == (char *) 0) ? infile : open (infile_name, O_RDONLY)) >= 0) /* open the input file */
        {
            retval = EOPENOUT; /* assume error opening output file */

            if ((outfile = (outfile_name == (char *) 0) ? outfile : open (outfile_name, O_WRONLY | O_CREAT, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH)) >= 0) /* open the output file */
            {
                outlock.l_type = F_WRLCK; /* lock EOF of the output file for write */
                retval = EOLCK; /* assume error locking output file */

                if (fcntl (outfile, F_SETLKW, &outlock) == 0) /* write lock the EOF of the output file */
                {
                    retval = NOERROR; /* assume no error */

                    while ((transaction = *transaction_sequence) != '\0') /* the transaction sequence that will be followed */
                    {

                        switch (transaction) /* which transaction? */
                        {

                            case 'r': /* read? */
                            case 'R':

                                retval = file_operation (interin, outfile, argc, argv, arg_ctr, 'r');
                                break;

                            case 'w': /* write? */
                            case 'W':

                                retval = file_operation (infile, interout, argc, argv, arg_ctr, 'a');
                                break;

                            default:

                                break;

                        }

                        transaction_sequence ++; /* increment the transaction sequence that will be followed */
                    }

                    if (add_eot == 1) /* add EOT = ^D = end of transmission/transaction marker character to output file, 0 = no, 1 = yes? */
                    {

                        if (write (outfile, &EOT, (size_t) 1) != (ssize_t) 1) /* write the EOT character to the output file */
                        {

                            if (retval == NOERROR) /* any previous errors? */
                            {
                                retval = EWROUT; /* assume error writing to output file */
                            }

                        }

                    }

                    outlock.l_type = F_UNLCK; /* unlock the file */

                    if (fcntl (outfile, F_SETLKW, &outlock) != 0) /* unlock the input file of the parent_write() function */
                    {

                        if (retval == NOERROR) /* any previous errors? */
                        {
                            retval = EOULCK; /* assume error unlocking file */
                        }

                    }

                }

                if (outfile != STDOUT_FILENO) /* output file open? */
                {

                    if (close (outfile) != 0) /* close the output file */
                    {

                        if (retval == NOERROR) /* any previous errors? */
                        {
                            retval = ECLOSEOUT; /* assume error closing output file */
                        }

                    }

                    outfile = STDOUT_FILENO; /* reset output file descriptor */
                }

            }

            if (infile != STDIN_FILENO) /* input file open? */
            {

                if (close (infile) != 0) /* close the input file */
                {

                    if (retval == NOERROR) /* any previous errors? */
                    {
                        retval = ECLOSEIN; /* assume error closing input file */
                    }

                }

                infile = STDIN_FILENO; /* reset input file descriptor */
            }

        }

    }

    print_message (retval); /* print any error messages */
    exit (retval); /* exit with the error value */

#ifdef LINT

    return (0); /* for lint formalities */

#endif

}

/*

File i/o operation.

static int file_operation (int read_in, int write_out, int argc, char *argv[], int arg_ctr, char eot_op);

I) Operation:

    A) If optind > argc, (argc and argv are command line arguments,)
       write the command line arguments to file descriptor write_out
       as EOL = '\n' terminated character strings.

    B) If optind <= argc, read bytes from file descriptor read_in, and
       write to file descriptor write_out.

II) Execution:

    A) The two files should be open: read_in should be open for at
       least reading; write_out should be open for at least writing.

    B) There are two EOT marker operations based on the eot_op:

        1) if eot_op is a 'a' character:

            a) at EOF, an EOT character is added to the output stream

            b) the read/write loop is exited

        2) if eot_op is a 'r' character:

            a) at EOT, the EOT character is removed from the input
               stream

            b) the read/write loop is exited

III) File desciptors, fd:

    A) When operating as a server under tcpserver:

        1) interin = fd 0, (data from the network)

        2) interout = fd 1, (data to the network)

        3) infile defaults to stdin, (data from a device/file)

        4) outfile defaults to stdout, (data to a device/file)

    B) When operating as a client under tcpclient:

        1) interin = fd 6, (data from the network)

        2) interout = fd 7, (data to the network)

        3) infile defaults to stdin, (data from a device/file)

        4) outfile defaults to stdout, (data to a device/file)

IV) Logic of transaction marker:

    A) When operating as a server under tcpserver:

        1) -s w, infile -> interout, if no EOT, add EOT at EOF, else
                                     EOT marks EOF

        2) -s r, interin -> outfile, remove EOT, (EOT must exist, wait
                                     for it,) which marks EOF

    B) When operating as a client under tcpclient:

        1) -c w, infile -> interout, if no EOT, add EOT at EOF, else
                                     EOT marks EOF

        2) -c r, interin -> outfile, remove EOT, (EOT must exist, wait
                                     for it,) which marks EOF

    Which is the same for -c and/or -s, and the same for -c w and -s
    w, and -c r and -s r.

V) Combining III and IV, the logic simplifies to:

    A) if operating as 'w':

        file_operation (infile, interout, argc, argv, arg_ctr, 'a');

    B) else if operating as 'r':

        file_operation (interin, outfile, argc, argv, arg_ctr, 'r');

    And, it doesn't make any difference whether it is operaating as a
    client under tcpclient, or server, under tcpserver, where 'a'
    means append an EOT character, and 'r' means remove the EOT
    character after, (possibly,) waiting for it.

Returns NOERROR on success, or a positive error representative of
failure.

*/

#ifdef __STDC__

static int file_operation (int read_in, int write_out, int argc, char *argv[], int arg_ctr, char eot_op)

#else

static int file_operation (read_in, write_out, argc, argv, optind, eot_op)
int read_in;
int write_out;
int argc;
char *argv[];
int optind;
char eot_op;

#endif

{
    static int args_sent = 0; /* command line arguments sent, 0 = no, 1 = yes */

    char buffer[BUFLEN + 1], /* i/o buffer */
         *sentinel = &buffer[BUFLEN], /* reference to sentinel character in i/o buffer */
         *arg_ref,  /* reference to argument */
         *eof; /* reference to EOF in buffer */

    int retval = NOERROR, /* return value, assume no error */
        quit_flag = 0, /* flag to stop reading read_in after current write to write_out, 0 = no, 1 = yes */
        i; /* argument counter */

    size_t len; /* length of character string */

    ssize_t cnt = (ssize_t) -1; /* number of bytes read from input file */

    *sentinel = '\0'; /* i/o buffer sentinel */

    if ((args_sent == 0) && (argc - arg_ctr > 0)) /* input as command line arguments? */
    {

        for (i = arg_ctr; (retval == NOERROR) && (quit_flag == 0) && (i < argc); i ++) /* for each argument */
        {
            arg_ref = argv[i];  /* reference the argument */
            len = strlen (arg_ref); /* save the length of the argument */
            retval = EWROUT; /* assume error writing to output file */

            if ((eof = strchr (arg_ref, EOT)) != (char *) 0) /* argument contain an EOT character? */
            {
                len = eof - buffer; /* adjust the number of bytes in the argument */
                quit_flag = 1; /* schedule a stop after current argument is written to write_out */
            }

            if (write (write_out, arg_ref, len) == (ssize_t) len) /* write the argument to the output file */
            {

                if (write (write_out, "\n", (size_t) 1) == (ssize_t) 1)
                {
                    retval = NOERROR; /* assume no error */
                }

            }

            args_sent = 1; /* set the command line arguments sent, 0 = no, 1 = yes */
        }

        if ((eot_op == 'a') && (retval == NOERROR)) /* need to add an EOT and no errors? */
        {
            retval = EWROUT; /* assume error writing to output file */

            if (write (write_out, &EOT, (size_t) 1) == (ssize_t) 1) /* write the EOT character to the output file */
            {
                retval = NOERROR; /* assume no error */
            }

        }

    }

    else
    {

        while ((retval == NOERROR) && (quit_flag == 0) && ((cnt = read (read_in, buffer, (size_t) BUFLEN)) >= (ssize_t) 0)) /* read characters from the input file while no errors, a quit scheduled, or EOF */
        {

            if (cnt == 0) /* any characters? */
            {

                if (eot_op == 'r') /* need to remove EOT character and quit if any exist in the buffer? */
                {
                    sleep (1); /* wait one second for the data to be appended to the output file which might have an EOT character */
                    continue; /* next read */
                }

                else
                {
                    break; /* finished */
                }

            }

            else /* something to write */
            {

                if (eot_op == 'r') /* need to remove EOT character and quit if any exist in the buffer? */
                {
                    buffer[cnt] = '\0'; /* terminate the characters with an EOS, (if the buffer is full, this will overwrite the existing EOS sentinel) */

                    if ((eof = strchr (buffer, EOT)) != (char *) 0) /* buffer contain an EOT character? */
                    {
                        *eof = '\0'; /* remove the EOT character */
                        cnt = eof - buffer; /* adjust the number of bytes read from input file to exclude the EOT marker */
                        quit_flag = 1; /* schedule a stop after current buffer is written to write_out */
                    }

                }

                retval = EWROUT; /* assume error writing to output file */

                if (write (write_out, buffer, (size_t) cnt) == cnt) /* write the characters to the output file */
                {
                    retval = NOERROR; /* assume no error */
                }

            }

        }

        if (cnt < (ssize_t) 0) /* read () error, (cnt == 0 is normal EOF, cnt == -1, a read error)? */
        {
            retval = ERDIN; /* assume error reading input file */
        }

        if ((eot_op == 'a') && (retval == NOERROR)) /* need to add an EOT and no errors? */
        {
            retval = EWROUT; /* assume error writing to output file */

            if (write (write_out, &EOT, (size_t) 1) == (ssize_t) 1) /* write the EOT character to the output file */
            {
                retval = NOERROR; /* assume no error */
            }

        }

    }

    return (retval); /* return an errors */
}

/*

Print any error messages.

static void print_message (int retval);

I) Data structures:

    A) The help_message array is an array of character stings, one per
       line to be printed for help.

    B) The error_message array is an array of character strings, one
       line per error; the error_message array is implicitly addressed
       by the value integer retval, which specifies the error to be
       printed.

II) Function execution:

    A) Depending on the value of the integer retval:

        1) If retval is zero, print nothing-normal/successful program
           exit.

        2) Else if retval is unity, print help.

        3) Else retval is not zero or unity, it is an error code, print
           the corresponding error message.

Returns nothing.

*/

#ifdef __STDC__

static void print_message (int retval)

#else

static void print_message (retval)
int retval;

#endif

{
    size_t help_ctr; /* help_message line counter */

    switch (retval) /* which return value */
    {

        case 0: /* program ended without errors, print nothing */

            break;

        case 1: /* program ended with a request for help, or argument error, print help */

            for (help_ctr = 0; help_ctr < (sizeof (help_message) / sizeof (char *)); help_ctr ++) /* for each line of help */
            {
                (void) printf ("%s", help_message[help_ctr]); /* print the line */
            }

            break;

        default: /* an error that was not a request for help, print the error */

            (void) fprintf (stderr, "%s", error_message[retval]);
            break;

    }

}
